//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2012 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;

using Smdn.IO;
using Smdn.Formats;
using Smdn.Formats.Mime;

namespace Smdn.Applications.MailReassembler {
  class MainClass {
    private static bool dryRun = false;

    public static void Main(string[] args)
    {
      string directory = null;
      string outputFilename = null;
      bool extractAttachedFiles = false;

      for (var i = 0; i < args.Length; i++) {
        switch (args[i]) {
          case "/f":
            outputFilename = args[++i];
            break;

          case "/?":
          case "/help":
          case "-h":
          case "--help":
            Usage(null);
            break;

          case "/e":
            extractAttachedFiles = true;
            break;

          case "/n":
            dryRun = true;
            break;

          default:
            if (Directory.Exists(args[i]))
              directory = args[i];
            break;
        }
      }

      if (directory == null) {
        Usage("結合したい分割メールが保存されているディレクトリを指定してください。");
        return;
      }

      var reassembled = Reassemble(LoadMessages(directory));

      Save(reassembled, directory, outputFilename);

      if (!dryRun && extractAttachedFiles)
        ExtractAttachedFile(reassembled, directory);
    }

    private static void Usage(string format, params object[] args)
    {
      var exitCode = 0;

      Console.WriteLine("使い方: {0} [オプション] ディレクトリ", Path.GetFileName(Assembly.GetEntryAssembly().CodeBase));
      Console.WriteLine();
      Console.WriteLine("指定されたディレクトリにある分割メール(emlファイル)を結合して保存します。");
      Console.WriteLine();
      Console.WriteLine("オプション:");
      Console.WriteLine("  /f ファイル名 : 結合したメールを保存する際のファイル名を指定します。");
      Console.WriteLine("                 省略した場合は、メールの件名もしくはMessage-IDが使われます。");
      Console.WriteLine("  /e           : 結合したメールにファイルが添付されている場合、添付ファイルを");
      Console.WriteLine("                 ディレクトリに保存します。");
      Console.WriteLine("  /n           : 結合できるかテストを行い、結合したメールの保存は行いません。");
      Console.WriteLine("  /?, /help    : 使い方を表示します。");
      Console.WriteLine();

      if (format != null) {
        Console.Error.WriteLine(format, args);
        exitCode = 1;
      }

      Environment.Exit(exitCode);
    }

    private static IEnumerable<IMimeMessage> LoadMessages(string directory)
    {
      Console.WriteLine("ディレクトリ '{0}' を検索しています ...", directory);

      try {
        return Directory.GetFiles(directory, "*.eml", SearchOption.TopDirectoryOnly).Select(Load).ToArray();
      }
      catch (EncodingNotSupportedException ex) {
        throw new ApplicationException(string.Format("デコードできない文字コード '{0}' が使われています。", ex.EncodingName), ex);
      }
    }

    private static IMimeMessage Load(string path)
    {
      Console.WriteLine("'{0}' を読み込んでいます ...", Path.GetFileName(path));

      return MimeMessage.Load(path);
    }

    private static MimeMessage Reassemble(IEnumerable<IMimeMessage> messages)
    {
      Console.WriteLine("メールを結合しています ...");

      try {
        try {
          return MimeMessage.ReassembleFragmentedMessage(messages);
        }
        catch (ArgumentException ex) {
          throw new ApplicationException("結合に必要なメールが欠けています。", ex);
        }
      }
      finally {
        foreach (var message in messages) {
          var disposable = message as IDisposable;

          if (disposable != null)
            disposable.Dispose();
        }
      }
    }

    private static void Save(MimeMessage message, string directory, string path)
    {
      if (!dryRun)
        Console.WriteLine("結合したメールを保存しています ...");

      if (path == null) {
        // set default filename
        string filename = null;

        if (message.Headers.Contains("Subject"))
          filename = message.Headers["Subject"].Value;
        else if (message.Headers.Contains("Message-ID"))
          filename = message.Headers["Message-ID"].Value;
        else
          filename = string.Empty;

        filename = PathUtils.RemoveInvalidFileNameChars(filename);

        if (filename.Length == 0)
          filename = "reassembled";

        path = Path.Combine(directory, filename + ".eml");
      }

      if (dryRun) {
        Console.WriteLine("結合したメールは '{0}' に保存されます。", path);
        return;
      }

      if (File.Exists(path)) {
        Console.Write("ファイル '{0}' が既に存在します。　上書きしますか? (y/N) ", path);

        var yesNo = Console.ReadLine();

        if (yesNo == null || !yesNo.StartsWith("y", StringComparison.OrdinalIgnoreCase)) {
          Console.WriteLine("保存しませんでした。");
          return;
        }
        else {
          File.Delete(path);
        }
      }

      message.Save(path);

      Console.WriteLine("結合したメールを '{0}' に保存しました。", path);
    }

    private static void ExtractAttachedFile(MimeMessage message, string directory)
    {
      Console.WriteLine("添付されているファイルを保存しています ...");

      var mail = new Mail(message);
      var files = mail.SaveAttachedFilesToDirectory(directory);

      if (0 < files.Length) {
        Console.WriteLine("以下の添付ファイルを保存しました。");

        foreach (var file in files) {
          Console.WriteLine("  {0}", file.FullName);
        }
      }
    }
  }
}
